import re
import uuid

from antlr4 import *
from antlr4.error.ErrorListener import ErrorListener
from dateutil import parser as dateparser
from textwrap import dedent

from norm import config
from norm.executable import NormExecutable, Projection
from norm.executable.command import Command
from norm.executable.constant import Constant
from norm.executable.schema.declaration import *
from norm.executable.expression.arithmetic import *
from norm.executable.expression.evaluation import *
from norm.executable.expression.query import *
from norm.executable.expression.slice import *
from norm.executable.schema.implementation import *
from norm.executable.schema.type import *
from norm.executable.schema.namespace import *
from norm.grammar.literals import AOP, COP, LOP, ImplType, ConstantType, MOP
from norm.grammar.normLexer import normLexer
from norm.grammar.normListener import normListener
from norm.grammar.normParser import normParser


class ParseError(ValueError):
    pass


class NormErrorListener(ErrorListener):

    def __init__(self):
        super(NormErrorListener, self).__init__()

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        err_msg = "line " + str(line) + ":" + str(column) + " " + msg
        raise ValueError(err_msg)


walker = ParseTreeWalker()


class NormCompiler(normListener):
    TMP_VARIABLE_STUB = 'tmp_'

    def __init__(self, context_id, user, session=None):
        self.context_id = context_id
        self.scopes = []
        self.stack = []
        self.session = session
        self.user = user
        self.context_namespace = None
        self.user_namespace = None
        self.search_namespaces = None
        self.set_namespaces()

    def set_namespaces(self):
        if self.context_id:
            self.context_namespace = '{}.{}'.format(config.CONTEXT_NAMESPACE_STUB, self.context_id)
        if self.user:
            self.user_namespace = '{}.{}'.format(config.USER_NAMESPACE_STUB, self.user.username)
        from norm.models import NativeLambda, CoreLambda
        self.search_namespaces = [NativeLambda.NAMESPACE, CoreLambda.NAMESPACE, self.context_namespace,
                                  self.user_namespace]

    def set_temp_scope(self):
        """
        For a unnamed query, we assign a temporary type for the scope.
        For a named query, i.e., type implementation, the scope is the type.
        """
        self.scopes.append((Lambda(self.context_namespace, self.TMP_VARIABLE_STUB + str(uuid.uuid4())), 'temp'))

    def get_scope(self, name):
        for scope, scope_lex in reversed(self.scopes):
            if name in scope:
                return scope
        return None

    @property
    def scope(self):
        return self.scopes[-1][0] if len(self.scopes) > 0 else None

    @property
    def scope_lex(self):
        return self.scopes[-1][1] if len(self.scopes) > 0 else None

    def _push(self, exe):
        self.stack.append(exe)

    def _pop(self):
        """
        :rtype: NormExecutable
        """
        return self.stack.pop()

    def _peek(self):
        """
        :rtype: NormExecutable
        """
        return self.stack[-1] if len(self.stack) > 0 else None

    def optimize(self, exe):
        """
        Optimize the AST to have a more efficient execution plan
        # TODO: optimization strategies
        * Filtering conditions can be combined and executed in batch instead of sequential
        * Arithmetic equations can be combined and passed to DF in batch instead of sequential
        :param exe: the executable to be optimized
        :type exe: NormExecutable
        :return: the optimized executable
        :rtype: NormExecutable
        """
        return exe

    def compile(self, script):
        if script is None or not isinstance(script, str):
            return None
        script = script.strip(' \r\n\t')
        if script == '':
            return None

        lexer = normLexer(InputStream(script))
        stream = CommonTokenStream(lexer)
        parser = normParser(stream)
        parser.addErrorListener(NormErrorListener())
        tree = parser.script()
        walker.walk(self, tree)
        # assert(len(self.stack) == 1)  # Ensure that parsing and compilation has finished completely
        # return self.optimize(self._pop())
        return

    def execute(self, script):
        self.stack = []
        self.scopes = []
        self.compile(dedent(script))
        results = None
        while len(self.stack) > 0:
            exe = self._pop()
            if isinstance(exe, NormExecutable):
                results = exe.execute(self)
            else:
                # TODO: shouldn't be here?
                results = exe
            if isinstance(results, Index) and isinstance(exe, NormExpression) and exe.lam is not None:
                results = exe.lam.data.loc[results]
        if isinstance(results, DataFrame):
            fix_dot_columns = OrderedDict()
            for col in results.columns:
                if col.find(NormExecutable.VARIABLE_SEPARATOR) >= 0:
                    fix_dot_columns[col] = col.replace(NormExecutable.VARIABLE_SEPARATOR, '.')
            results = results.rename(columns=fix_dot_columns)
        return results

    def exitStatement(self, ctx: normParser.StatementContext):
        if ctx.typeDeclaration():
            if ctx.IMPL():
                query = self._pop()  # type: NormExpression
                type_declaration = self._pop()  # type: TypeDeclaration
                description = self._pop() if ctx.comments() else ''
                type_declaration.description = description
                op = ImplType(ctx.IMPL().getText())
                self._push(TypeImplementation(type_declaration, op, query, description).compile(self))
            elif ctx.comments():
                # otherwise just pass through
                type_declaration = self._pop()  # type: TypeDeclaration
                description = self._pop()
                type_declaration.description = description  # need to be recompiled
                self._push(type_declaration.compile(self))
        elif ctx.argumentDeclarations():
            args = self._pop()  # type: List[ArgumentDeclaration]
            type_ = self._pop()  # type: TypeName
            self._push(AdditionalTypeDeclaration(type_, args).compile(self))
        elif ctx.renames():
            renames = self._pop()  # type: List[RenameArgument]
            type_ = self._pop()  # type: TypeName
            self._push(RenameTypeDeclaration(type_, renames).compile(self))
        elif ctx.codeExpression():
            code = self._pop()  # type: str
            type_ = self._pop()  # type: TypeName
            description = self._pop() if ctx.comments() else ''
            self._push(CodeTypeDeclaration(type_, code, description).compile(self))
        elif ctx.imports() or ctx.exports() or ctx.commands() or ctx.multiLineExpression():
            if ctx.comments():
                expr = self._pop()
                # ignore comments
                self._pop()
                self._push(expr)

    def exitNone(self, ctx: normParser.NoneContext):
        self._push(Constant(ConstantType.NULL, None))

    def exitBool_c(self, ctx: normParser.Bool_cContext):
        self._push(Constant(ConstantType.BOOL, ctx.getText().lower() == 'true'))

    def exitInteger_c(self, ctx: normParser.Integer_cContext):
        self._push(Constant(ConstantType.INT, int(ctx.getText())))

    def exitFloat_c(self, ctx: normParser.Float_cContext):
        self._push(Constant(ConstantType.FLT, float(ctx.getText())))

    def exitString_c(self, ctx: normParser.String_cContext):
        self._push(Constant(ConstantType.STR, str(ctx.getText()[1:-1])))

    def exitPattern(self, ctx: normParser.PatternContext):
        try:
            self._push(Constant(ConstantType.PTN, re.compile(str(ctx.getText()[2:-1]))))
        except:
            raise ParseError('Pattern constant {} is in wrong format, should be Python regex pattern'
                             .format(ctx.getText()))

    def exitUuid(self, ctx: normParser.UuidContext):
        self._push(Constant(ConstantType.UID, str(ctx.getText()[2:-1])))

    def exitUrl(self, ctx: normParser.UrlContext):
        self._push(Constant(ConstantType.URL, str(ctx.getText()[2:-1])))

    def exitDatetime(self, ctx: normParser.DatetimeContext):
        self._push(Constant(ConstantType.DTM, dateparser.parse(ctx.getText()[2:-1], fuzzy=True)))

    def exitConstant(self, ctx: normParser.ConstantContext):
        if ctx.LSBR():
            constants = list(reversed([self._pop() for ch in ctx.children
                                       if isinstance(ch, normParser.ConstantContext)]))  # type: List[Constant]
            types = set(constant.type_ for constant in constants)
            if len(types) > 1:
                type_ = ConstantType.ANY
            else:
                type_ = types.pop()
            self._push(ListConstant(type_, [constant.value for constant in constants]))

    def exitQueryProjection(self, ctx: normParser.QueryProjectionContext):
        variables = list(reversed([self._pop() for ch in ctx.children
                                   if isinstance(ch, normParser.VariableContext)]))
        to_evaluate = True if ctx.LCBR() else False
        # TODO: evaluate the variables to get the referred variables
        self._push(Projection(variables, to_evaluate))

    def exitComments(self, ctx: normParser.CommentsContext):
        spaces = ' \r\n\t'
        cmt = ctx.getText()
        if ctx.MULTILINE():
            cmt = cmt.strip(spaces)[2:-2].strip(spaces)
        elif ctx.SINGLELINE():
            cmt = '\n'.join(cmt_line.strip(spaces)[2:].strip(spaces) for cmt_line in cmt.split('\n'))
        self._push(cmt)

    def exitImports(self, ctx: normParser.ImportsContext):
        type_: TypeName = self._pop() if ctx.typeName() else None
        namespace = [str(v) for v in ctx.VARNAME()]
        variable = namespace.pop() if ctx.AS() else None
        self._push(Import('.'.join(namespace), type_, variable).compile(self))

    def exitExports(self, ctx: normParser.ExportsContext):
        type_: TypeName = self._pop()
        namespace = [str(v) for v in ctx.VARNAME()]
        variable = namespace.pop() if ctx.AS() else None
        self._push(Export('.'.join(namespace), type_, variable).compile(self))

    def exitCommands(self, ctx: normParser.CommandsContext):
        type_: TypeName = self._pop()
        op = MOP(ctx.SPACED_COMMAND().getText().strip().lower())
        self._push(Command(op, type_).compile(self))

    def exitArgumentDeclaration(self, ctx: normParser.ArgumentDeclarationContext):
        variable_property = ctx.argumentProperty().getText() if ctx.argumentProperty() else None
        optional = variable_property is not None and variable_property.lower().find('optional') > 0
        type_name: TypeName = self._pop()
        variable_name: VariableName = self._pop()
        self._push(ArgumentDeclaration(variable_name, type_name, optional))

    def enterArgumentDeclarations(self, ctx: normParser.ArgumentDeclarationsContext):
        type_name = self._peek()
        if type_name is not None and isinstance(type_name, TypeName) and type_name.lam is not None:
            self.scopes.append((type_name.lam, 'argument_declarations'))

    def exitArgumentDeclarations(self, ctx: normParser.ArgumentDeclarationsContext):
        args = list(reversed([self._pop() for ch in ctx.children
                              if isinstance(ch, normParser.ArgumentDeclarationContext)]))
        self._push(args)
        if self.scope_lex == 'argument_declarations':
            self.scopes.pop()

    def exitRename(self, ctx: normParser.RenameContext):
        new_name: VariableName = self._pop()
        original_name: VariableName = self._pop()
        self._push(RenameArgument(original_name.name, new_name.name))

    def exitRenames(self, ctx: normParser.RenamesContext):
        args = list(reversed([self._pop() for ch in ctx.children
                              if isinstance(ch, normParser.RenameContext)]))
        self._push(args)
        if self.scope_lex == 'renames':
            self.scopes.pop()

    def enterRenames(self, ctx: normParser.RenamesContext):
        type_name = self._peek()
        if type_name is not None and isinstance(type_name, TypeName) and type_name.lam is not None:
            self.scopes.append((type_name.lam, 'renames'))

    def exitTypeDeclaration(self, ctx: normParser.TypeDeclarationContext):
        output_type_name = self._pop() if ctx.typeName(1) else None  # type: TypeName
        args = self._pop() if ctx.argumentDeclarations() else None  # type: List[ArgumentDeclaration]
        type_name = self._pop()  # type: TypeName
        self._push(TypeDeclaration(type_name, args, output_type_name).compile(self))

    def exitTypeName(self, ctx: normParser.TypeNameContext):
        typename = ctx.VARNAME()
        if typename:
            version = ctx.version().getText() if ctx.version() else None
            self._push(TypeName(str(typename), version).compile(self))
        elif ctx.LSBR():
            intern = self._pop()  # type: TypeName
            self._push(ListType(intern).compile(self))
        else:
            raise ParseError('Not a valid type name definition')

    def exitVariable(self, ctx: normParser.VariableContext):
        name: str = ''
        if ctx.VARNAME():
            name = ctx.VARNAME().getText()
        elif ctx.COMMAND():
            name = ctx.COMMAND().getText()
        elif ctx.ARGOPT():
            name = ctx.ARGOPT().getText()
        scope = self._pop() if ctx.variable() else None  # type: VariableName
        self._push(VariableName(scope, name).compile(self))

    def exitArgumentExpression(self, ctx: normParser.ArgumentExpressionContext):
        if isinstance(self._peek(), ArgumentExpr):
            return

        projection = self._pop() if ctx.queryProjection() else None  # type: Projection
        expr = self._pop() if ctx.arithmeticExpression() else None  # type: ArithmeticExpr
        op = COP(ctx.spacedConditionOperator().conditionOperator().getText().lower()) \
            if ctx.spacedConditionOperator() else None  # type: COP
        # = is treated as op is None
        variable = self._pop() if ctx.variable() else None  # type: VariableName
        if variable is None and projection is None and op is None and expr is not None \
                and isinstance(expr, EvaluationExpr) and isinstance(expr.variable, ColumnVariable) \
                and len(expr.args) == 0 and expr.projection is not None:
            self._push(ArgumentExpr(expr.variable, op, None, expr.projection).compile(self))
        else:
            self._push(ArgumentExpr(variable, op, expr, projection).compile(self))

    def enterArgumentExpressions(self, ctx: normParser.ArgumentExpressionsContext):
        expr = self._peek()
        if expr is not None and isinstance(expr, (VariableName, EvaluationExpr)) and expr.lam is not None:
            self.scopes.append((expr.lam, 'argument_expressions'))

    def exitArgumentExpressions(self, ctx: normParser.ArgumentExpressionsContext):
        args = list(reversed([self._pop() for ch in ctx.children
                              if isinstance(ch, normParser.ArgumentExpressionContext)]))
        self._push(args)
        if self.scope_lex == 'argument_expressions':
            self.scopes.pop()

    def enterMultiLineExpression(self, ctx: normParser.MultiLineExpressionContext):
        type_declaration = self._peek()
        if type_declaration is not None and isinstance(type_declaration, TypeDeclaration) \
                and type_declaration.lam is not None:
            self.scopes.append((type_declaration.lam, 'multiline'))

    def exitContext(self, ctx:normParser.ContextContext):
        type_name = self._pop()
        self.scopes.append((type_name.lam, 'multiline'))

    def exitMultiLineExpression(self, ctx: normParser.MultiLineExpressionContext):
        if ctx.newlineLogicalOperator():
            expr2 = self._pop()  # type: NormExpression
            expr1 = self._pop()  # type: NormExpression
            op = LOP.parse(ctx.newlineLogicalOperator().logicalOperator().getText())
            self._push(QueryExpr(op, expr1, expr2).compile(self))
        if self.scope_lex == 'multiline':
            self.scopes.pop()

    def enterSpacedLogicalOperator(self, ctx:normParser.SpacedLogicalOperatorContext):
        expr = self._peek()
        self.scopes.append((expr.lam, 'lop'))

    def exitOneLineExpression(self, ctx: normParser.OneLineExpressionContext):
        if ctx.queryProjection():
            projection = self._pop()
            expr = self._peek()
            expr.projection = projection
            expr.compile(self)
        elif ctx.NOT():
            expr = self._pop()  # type: NormExpression
            self._push(NegatedQueryExpr(expr).compile(self))
        elif ctx.spacedLogicalOperator():
            expr2 = self._pop()  # type: NormExpression
            expr1 = self._pop()  # type: NormExpression
            op = LOP.parse(ctx.spacedLogicalOperator().logicalOperator().getText())
            self._push(QueryExpr(op, expr1, expr2).compile(self))
            if self.scope_lex == 'lop':
                self.scopes.pop()

    def exitConditionExpression(self, ctx: normParser.ConditionExpressionContext):
        if ctx.spacedConditionOperator():
            qexpr = self._pop()  # type: ArithmeticExpr
            aexpr = self._pop()  # type: ArithmeticExpr
            cop = COP(ctx.spacedConditionOperator().conditionOperator().getText().lower())
            self._push(ConditionExpr(cop, aexpr, qexpr).compile(self))

    def exitArithmeticExpression(self, ctx: normParser.ArithmeticExpressionContext):
        if ctx.slicedExpression():
            return

        op = None
        if ctx.MOD():
            op = AOP.MOD
        elif ctx.EXP():
            op = AOP.EXP
        elif ctx.TIMES():
            op = AOP.MUL
        elif ctx.DIVIDE():
            op = AOP.DIV
        elif ctx.PLUS():
            op = AOP.ADD
        elif ctx.MINUS():
            op = AOP.SUB

        if op is not None:
            expr2 = self._pop()  # type: ArithmeticExpr
            if isinstance(expr2, ArgumentExpr):
                expr2 = expr2.variable
            expr1 = self._pop() if ctx.arithmeticExpression(1) else None  # type: ArithmeticExpr
            if isinstance(expr1, AddDataEvaluationExpr):
                expr1 = list(expr1.data.values())[0]
                assert (isinstance(expr1, ArithmeticExpr))
            self._push(ArithmeticExpr(op, expr1, expr2).compile(self))

    def exitSlicedExpression(self, ctx: normParser.SlicedExpressionContext):
        if ctx.LSBR():
            if ctx.evaluationExpression(1):
                expr_range = self._pop()  # type: NormExpression
                expr = self._pop()  # type: NormExpression
                self._push(EvaluatedSliceExpr(expr, expr_range).compile(self))
            else:
                end = self._pop() if ctx.integer_c(1) else None  # type: Constant
                start = self._pop() if ctx.integer_c(0) else None  # type: Constant
                expr = self._pop()
                start_value = start.value
                if end is not None:
                    end_value = end.value
                elif ctx.COLON():
                    colon_before_number = True
                    for ch in ctx.children:
                        if ch == ctx.COLON():
                            break
                        elif ch == ctx.integer_c(0):
                            colon_before_number = False
                            break
                    if colon_before_number:
                        end_value = start_value
                        start_value = 0
                    else:
                        end_value = None
                else:
                    end_value = start_value + 1
                self._push(SliceExpr(expr, start_value, end_value).compile(self))

    def exitEvaluationExpression(self, ctx: normParser.EvaluationExpressionContext):
        if ctx.DOT():
            rexpr = self._pop()
            lexpr = self._pop()
            self._push(ChainedEvaluationExpr(lexpr, rexpr).compile(self))
        elif ctx.argumentExpressions() or ctx.queryProjection():
            projection = self._pop() if ctx.queryProjection() else None
            args = self._pop() if ctx.argumentExpressions() else []  # type: List[ArgumentExpr]
            variable = self._pop() if ctx.variable() else None  # type: VariableName
            self._push(EvaluationExpr(args, variable, projection).compile(self))

    def exitCodeExpression(self, ctx: normParser.CodeExpressionContext):
        self._push(ctx.code().getText())



from norm.executable.expression import NormExpression
from norm.executable.schema.variable import VariableName


class AssignmentExpr(NormExpression):

    def __init__(self, variable, expr):
        """
        Assigning the result of an expression to an object variable
        :param variable: the name of the variable
        :type variable: VariableName
        :param expr: the expression to be evaluated at
        :type expr: ArithmeticExpr
        """
        super().__init__()
        self.variable = variable
        self.expr = expr

    def serialize(self):
        pass

    def execute(self, context):
        # TODO: assign the value to the object variable
        pass



import uuid

from norm.utils import hash_df
from pandas import DataFrame, Series, Index

from norm.grammar.literals import COP
from norm.executable import NormError, Projection
from norm.executable.expression.argument import ArgumentExpr
from norm.executable.schema.variable import VariableName, ColumnVariable, JoinVariable
from norm.executable.expression import NormExpression

from collections import OrderedDict

from typing import List, Dict

import logging
logger = logging.getLogger(__name__)


class EvaluationExpr(NormExpression):

    def __init__(self, args, variable=None, projection=None):
        """
        The evaluation of an expression either led by a variable name
        :param args: the arguments provided
        :type args: List[ArgumentExpr]
        :param variable: the variable name to evaluate
        :type variable: VariableName
        :type projection: Projection
        """
        super().__init__()
        self.variable: VariableName = variable
        self.args: List[ArgumentExpr] = args
        self.inputs: Dict[str, DataFrame] = None
        self.outputs: Dict[str, str] = None
        self.join_variables: List[JoinVariable] = []
        self.projection: Projection = projection
        self.equalities: Dict[str, str] = None
        from norm.models.norm import Lambda
        self.equality_scope: Lambda = None

    @property
    def is_constant_arguments(self):
        return all(arg.is_constant for arg in self.args)

    def check_assignment_arguments(self):
        """
        Check whether the arguments are in the style of assignments. Ex. 234, 'asfs', b=3, c=5
        :return: True or False
        :rtype: bool
        """
        keyword_arg = False  # positional argument
        for arg in self.args:
            if arg.op is None:
                if arg.variable is None:
                    if keyword_arg:
                        msg = 'Keyword based arguments should come after positional arguments'
                        logger.error(msg)
                        logger.error([str(arg) for arg in self.args])
                        raise NormError(msg)
                else:
                    keyword_arg = True
            else:
                return False
        return True

    def build_assignment_inputs(self, lam):
        """
        If the arguments are in assignment style, build the inputs as assignments
        :param lam: given the Lambda to be evaluated
        :return: a dictionary mapping from a variable to an expression for the inputs
        :rtype: Dict
        """
        if lam.nargs == 0:
            # Lambda is anonymously created
            from norm.models.norm import Lambda
            return OrderedDict((arg.variable.name if arg.variable else '{}{}'.format(Lambda.VAR_ANONYMOUS_STUB, i),
                                arg.expr) for i, arg in enumerate(self.args))

        keyword_arg = False
        inputs = OrderedDict()
        from norm.models.norm import Variable
        for ov, arg in zip(lam.variables, self.args):  # type: Variable, ArgumentExpr
            if arg.expr is None:
                continue

            if arg.op is None and arg.variable is not None:
                keyword_arg = True
            if not keyword_arg:
                inputs[ov.name] = arg.expr
            else:
                inputs[arg.variable.name] = arg.expr
        return inputs

    def build_conditional_inputs(self):
        """
        If the arguments are in conditional style, build the inputs as conditionals
        :return: a query string
        :rtype: Dict
        """
        self.equalities = OrderedDict()
        self.equality_scope = None
        inputs = []
        for arg in self.args:
            if (arg.is_assign_operator or arg.op is COP.EQ) and isinstance(arg.expr, ColumnVariable) and \
                    isinstance(arg.variable, ColumnVariable):
                self.equalities[arg.expr.name] = arg.variable.name
                if self.equality_scope is None:
                    self.equality_scope = arg.expr.lam
                else:
                    assert(self.equality_scope is arg.expr.lam)
                continue
            if arg.op is None or arg.expr is None:
                continue
            vn = arg.variable.name
            if isinstance(arg.variable, JoinVariable):
                self.join_variables.append(arg.variable)
                vn = str(arg.variable)
            if arg.op == COP.LK:
                condition = '{}.str.contains({})'.format(vn, arg.expr)
            else:
                condition = '{} {} ({})'.format(vn, arg.op, arg.expr)
            inputs.append(condition)
        return ' and '.join(inputs)

    def build_outputs(self, lam):
        """
        Build the outputs according to the projection
        :return: a dictionary mapping from the original variable to the new variable
        :rtype: Dict
        """
        from norm.models.norm import Lambda
        assert(isinstance(lam, Lambda))
        outputs = OrderedDict()
        from norm.models.norm import Variable
        for ov, arg in zip(lam.variables, self.args):  # type: Variable, ArgumentExpr
            if arg.projection is not None:
                assert(len(arg.projection.variables) <= 1)
                assert(not arg.projection.to_evaluate)
                if arg.variable is None:
                    outputs[ov.name] = arg.projection.variables[0].name
                else:
                    if len(arg.projection.variables) == 0:
                        outputs[arg.variable.name] = arg.variable.name
                    else:
                        outputs[arg.variable.name] = arg.projection.variables[0].name
        if self.projection is not None:
            assert(len(self.projection.variables) <= 1)
            assert(not self.projection.to_evaluate)
            if self.projection.num == 1:
                new_lambda_name = self.projection.variables[0].name
            else:
                new_lambda_name = lam.name
            outputs = OrderedDict((key, self.VARIABLE_SEPARATOR.join([new_lambda_name, value])) for key, value
                                  in outputs.items())
            if lam.is_functional:
                outputs[lam.VAR_OUTPUT] = new_lambda_name
            else:
                outputs[lam.VAR_OID] = new_lambda_name
        return outputs

    def compile(self, context):
        if self.variable is None:
            from norm.models.norm import Lambda
            assert(context.scope is not None)
            assert(isinstance(context.scope, Lambda))
            lam = context.scope
            # constructing new objects
            assert(self.check_assignment_arguments())
            self.inputs = self.build_assignment_inputs(lam)
            self.outputs = None
            is_to_add_data = True
        else:
            lam = self.variable.lam
            if lam is None:
                # Might need to be combined with previous context
                return self
            if self.check_assignment_arguments():
                self.inputs = self.build_assignment_inputs(lam)
                is_to_add_data = not lam.atomic
            else:
                self.inputs = self.build_conditional_inputs()
                is_to_add_data = False
            self.outputs = self.build_outputs(lam)
            is_to_add_data &= len(self.outputs) == 0

        if self.equality_scope is not None and len(self.equalities) > 0:
            return JoinEqualityEvaluationExpr(lam, self.equality_scope, self.equalities, self.inputs, self.outputs)\
                .compile(context)

        if is_to_add_data:
            return AddDataEvaluationExpr(lam, self.inputs, self.variable is not None).compile(context)
        elif isinstance(self.inputs, dict):
            if self.projection is not None and self.projection.num == 1:
                output_projection = self.projection.variables[0].name
            else:
                output_projection = None
            if lam.atomic:
                return AtomicEvaluationExpr(lam, self.inputs, output_projection).compile(context)
            elif len(self.inputs) == 0:
                if isinstance(self.variable, ColumnVariable):
                    return ArgumentExpr(self.variable, None, None, self.projection).compile(context)
                elif self.projection is not None and len(self.outputs) == 1:
                    return RetrieveAllDataExpr(lam, output_projection).compile(context)
                else:
                    return RetrievePartialDataExpr(lam, self.outputs).compile(context)
        self.eval_lam = lam
        from norm.models import Lambda, Variable
        if len(self.outputs) > 0:
            variables = [Variable(self.outputs[v.name], v.type_) for v in self.eval_lam.variables
                         if v.name in self.outputs.keys()]
        else:
            variables = self.eval_lam.variables
        self.lam = Lambda(context.context_namespace, context.TMP_VARIABLE_STUB + str(uuid.uuid4()),
                          variables=variables)
        self.lam.cloned_from = self.eval_lam
        return self

    def execute(self, context):
        for jv in self.join_variables:
            jv.execute(context)
        df = self.eval_lam.data.query(self.inputs, engine='python')

        if isinstance(df, DataFrame) and len(self.outputs) > 0:
            df = df[self.outputs.keys()].rename(columns=self.outputs)
        self.lam.data = df
        return df


class JoinEqualityEvaluationExpr(NormExpression):

    def __init__(self, lam, scope, equalities, condition, outputs):
        """
        Join the input scope first and then evaluate
        Event(name?, ip?eip) & Cluster(ip=eip, name~'App'?cluster_name);
        :param lam: the current scope
        :param scope: the input scope
        :param equalities: the equalities, e.g., {'eip':'ip'}
        :param condition: the other condition, e.g., 'name.str.contains("App")'
        :param outputs: the outputs for the current scope, e.g., {'name':'cluster_name'}
        """
        super().__init__()
        self.eval_lam = lam
        self.scope = scope
        self.equalities = equalities
        self.condition = condition
        self.outputs = outputs

    def compile(self, context):
        scope = self.scope
        from norm.models import Lambda, Variable
        if len(self.outputs) > 0:
            variables = [Variable(v.name, v.type_) for v in scope.variables
                         if v.name in self.outputs.keys()]
            leftover = set(self.outputs.values()).difference(v.name for v in variables)
            variables += [Variable(self.outputs[v.name], v.type_) for v in self.eval_lam.variables
                          if self.outputs.get(v.name) in leftover]
        else:
            variables = scope.variables
            scope_variable_names = set(v.name for v in variables)
            variables += [v for v in self.eval_lam.variables if v.name not in scope_variable_names]
        self.lam = Lambda(context.context_namespace, context.TMP_VARIABLE_STUB + str(uuid.uuid4()),
                          variables=variables)
        self.lam.cloned_from = scope
        return self

    def execute(self, context):
        inp = self.lam.cloned_from.data
        equal_cols = list(self.equalities.items())
        to_merge = self.eval_lam.data.query(self.condition, engine='python')
        if len(self.outputs) > 0:
            to_merge = to_merge[set(list(self.outputs.keys()) + list(self.equalities.values()))]\
                        .rename(columns=self.outputs)
        left_col, right_col = equal_cols.pop()
        joined = inp.reset_index().merge(to_merge, left_on=left_col, right_on=right_col)
        joined = joined.set_index(self.lam.VAR_OID)
        condition = ' & '.join('({} == {})'.format(left_col, right_col) for left_col, right_col in equal_cols)
        if condition != '':
            results = joined.query(condition)
        else:
            results = joined
        self.lam.data = results
        return results


class AtomicEvaluationExpr(NormExpression):

    def __init__(self, lam, inputs, output_projection=None):
        """
        Evaluate the atomic function
        :param lam: the function to evaluate
        :param inputs: the input values
        :param output_projection: the output projection
        """
        super().__init__()
        from norm.models.norm import Lambda
        assert(lam is not None)
        assert(isinstance(lam, Lambda))
        assert(isinstance(inputs, dict))
        self.eval_lam: Lambda = lam
        self.lam: Lambda = lam.output_type
        self.inputs: Dict = inputs
        self.output_projection: str = output_projection

    def __str__(self):
        return '{}({})'.format(self.eval_lam.signature, ', '.join('{}={}'.format(key, value) for key, value
                                                                  in self.inputs.items()))

    def compile(self, context):
        # TODO: infer the output lambda schema
        return self

    def execute(self, context):
        inputs = dict((key, value.execute(context)) for key, value in self.inputs.items())
        result = self.eval_lam(**inputs)
        if self.output_projection is not None:
            import numpy
            if isinstance(result, (list, numpy.ndarray)):
                return Series(result, name=self.output_projection)
            elif isinstance(result, (Series, Index)):
                return result.rename(self.output_projection)
            elif isinstance(result, DataFrame):
                cols = {col: '{}{}{}'.format(self.output_projection, self.VARIABLE_SEPARATOR, col)
                        for col in result.columns}
                return result.loc[result.index.rename(self.output_projection)].rename(columns=cols)
            else:
                return Series([result], name=self.output_projection)
        return result


class RetrievePartialDataExpr(NormExpression):

    def __init__(self, lam, outputs):
        """
        Retrieve all data from the given Lambda
        :param lam: the given Lambda
        :param outputs: the output projections
        """
        super().__init__()
        from norm.models.norm import Lambda
        assert(lam is not None)
        assert(isinstance(lam, Lambda))
        assert(not lam.atomic)
        self.eval_lam: Lambda = lam
        self.lam: Lambda = None
        self.outputs: Dict = outputs

    def __str__(self):
        oid = self.eval_lam.VAR_OID
        out = self.eval_lam.VAR_OUTPUT
        return '{}({}){}'.format(self.eval_lam.signature,
                                 ', '.join(('{}?'.format(c) for c in self.outputs.keys() if c != oid and c != out)),
                                 '?' if self.outputs.get(oid) or self.outputs.get(out) else '')

    def compile(self, context):
        from norm.models import Lambda, Variable
        if len(self.outputs) > 0:
            variables = [Variable(self.outputs[v.name], v.type_) for v in self.eval_lam.variables
                         if v.name in self.outputs.keys()]
        else:
            variables = self.eval_lam.variables
        self.lam = Lambda(context.context_namespace, context.TMP_VARIABLE_STUB + str(uuid.uuid4()),
                          variables=variables)
        return self

    def execute(self, context):
        oid_output_col = self.outputs.get(self.eval_lam.VAR_OID)
        if oid_output_col is not None:
            del self.outputs[self.eval_lam.VAR_OID]

        result = self.eval_lam.data
        result = result[self.outputs.keys()].rename(columns=self.outputs)

        self.lam.data = result
        if oid_output_col is not None:
            result.index.rename(oid_output_col, inplace=True)
        return result


class RetrieveAllDataExpr(NormExpression):

    def __init__(self, lam, output_projection=None):
        """
        Retrieve all data from the given Lambda
        :param lam: the given Lambda
        :param output_projection: the output projection variable name
        """
        super().__init__()
        from norm.models.norm import Lambda
        assert(lam is not None)
        assert(isinstance(lam, Lambda))
        assert(not lam.atomic)
        self.eval_lam: Lambda = lam
        self.lam: Lambda = lam
        self.output_projection: str = output_projection

    def __str__(self):
        return '{}?'.format(self.lam.signature)

    def compile(self, context):
        return self

    def execute(self, context):
        if self.lam.is_functional:
            result = self.lam.data[self.lam.VAR_OUTPUT]
        else:
            result = self.lam.data.index
        if self.output_projection is not None:
            return result.rename(self.output_projection)
        else:
            return result


class AddDataEvaluationExpr(NormExpression):

    def __init__(self, lam, data, immediately=True):
        """
        Add data to a Lambda. If the lambda is not the current scope, revision occurs immediately. Otherwise,
        revision is delayed to TypeImplementation
        :param lam: the Lambda to add data
        :type lam: norm.model.norm.Lambda
        :param data: the data to add
        :type data: Dict
        :param immediately: whether to revise Lambda as a disjunction immediately, default to True
        :type immediately: bool
        """
        super().__init__()
        from norm.models.norm import Lambda
        self.lam: Lambda = lam
        self.data: Dict = data
        self.immediately: bool = immediately
        self.description: str = 'add data'
        self._query_str: str = None

    def __str__(self):
        if self._query_str is None:
            msg = "Need to execute first!"
            logger.error(msg)
            raise NormError(msg)
        return self._query_str

    def compile(self, context):
        if not self.immediately and len(self.data) == 1:
            expr = list(self.data.values())[0]
            from norm.executable.expression.arithmetic import ArithmeticExpr
            if isinstance(expr, ArithmeticExpr) and expr.op is not None:
                return expr
        return self

    def execute(self, context):
        if len(self.data) == 0:
            return DataFrame(data=[self.lam.default])

        data = OrderedDict((key, value.execute(context)) for key, value in self.data.items())
        data = self.unify(data)
        df = DataFrame(data=data, columns=data.keys())
        query_str = hash_df(df)
        self._query_str = query_str
        df = self.lam.fill_primary(df)
        df = self.lam.fill_time(df)
        df = self.lam.fill_oid(df)
        if self.lam.VAR_OID in df.columns:
            df = df.set_index(self.lam.VAR_OID)
        else:
            df.index.rename(self.lam.VAR_OID, inplace=True)
        if self.immediately:
            from norm.models.norm import RevisionType
            df = self.lam.revise(query_str, self.description, df, RevisionType.DISJUNCTION)
            if self.lam.is_functional:
                return df[self.lam.VAR_OUTPUT]
            else:
                return df.index
        return df


class ChainedEvaluationExpr(NormExpression):

    def __init__(self, lexpr, rexpr):
        """
        Chained evaluation expressions:
        Example:
        * a.b
        * a.test(sf, s)
        * a.b.c
        :param lexpr: base query expressions or chained expression
        :type lexpr: Union[VariableName, EvaluationExpr]
        :param rexpr: chained evaluation expression
        :type rexpr: Union[EvaluationExpr, VariableName]
        """
        super().__init__()
        self.lexpr = lexpr
        self.rexpr = rexpr

    def compile(self, context):
        if isinstance(self.rexpr, VariableName):
            if isinstance(self.lexpr, VariableName):
                return VariableName(self.lexpr, self.rexpr.name).compile(context)
            elif isinstance(self.lexpr, EvaluationExpr):
                if self.rexpr.name in self.lexpr.lam:
                    return ColumnVariable(self.lexpr, self.rexpr.name)
                else:
                    msg = '{} does not exist in the {}'.format(self.rexpr, self.lexpr)
                    logger.error(msg)
                    raise NormError(msg)
            else:
                msg = 'Only a variable or an evaluation can have chained operation for now, ' \
                      'but got {}'.format(self.lexpr)
                logger.error(msg)
                raise NormError(msg)
        elif isinstance(self.rexpr, EvaluationExpr):
            if self.rexpr.lam is None and isinstance(self.lexpr, ColumnVariable):
                return DataFrameColumnFunctionExpr(self.lexpr, self.rexpr).compile(context)
            else:
                # Evaluation with the previous expression as the first input argument
                self.rexpr.args = [ArgumentExpr(expr=self.lexpr)] + self.rexpr.args
                # Recompile the expression
                return self.rexpr.compile(context)
        else:
            msg = 'Only chaining on a variable or an evaluation, but got {}'.format(self.rexpr)
            logger.error(msg)
            raise NormError(msg)


class DataFrameColumnFunctionExpr(EvaluationExpr):

    def __init__(self, column_variable, expr):
        super().__init__([])
        self.column_variable: ColumnVariable = column_variable
        self.expr: EvaluationExpr = expr

    def compile(self, context):
        return self

    def execute(self, context):
        col = self.column_variable.execute(context)
        args = []
        kwargs = {}
        for arg in self.expr.args:  # type: ArgumentExpr
            if arg.expr is None:
                continue
            if arg.op is None and arg.variable is not None:
                kwargs[arg.variable.name] = arg.expr.execute(context)
            else:
                args.append(arg.expr.execute(context))
        func = self.expr.variable.name
        try:
            df = getattr(col, func)(*args, **kwargs)
        except:
            print('error on {}'.format(col))
            df = getattr(col.str, func)(*args, **kwargs)
            print(args, kwargs)
            print(df)
        return df



"""adding PythonLambda

Revision ID: 6939fca644fb
Revises: 467af4a6e6b3
Create Date: 2019-03-14 20:39:59.813682

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '6939fca644fb'
down_revision = '467af4a6e6b3'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('lambdas', sa.Column('code', sa.Text()))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('lambdas', 'code')
    # ### end Alembic commands ###



from sqlalchemy import Column, Integer, String, DateTime
from norm.models import Model


class User(Model):
    """
    Partial declaration to be compatible to the server side
    """

    __tablename__ = 'users'

    id = Column(Integer, primary_key=True)
    first_name = Column(String(64), nullable=False)
    last_name = Column(String(64), nullable=False)
    username = Column(String(64), unique=True, nullable=False)
    email = Column(String(64), unique=True, nullable=False)
    last_login = Column(DateTime)
    login_count = Column(Integer)
    fail_login_count = Column(Integer)

    @property
    def is_authenticated(self):
        return True

    @property
    def is_active(self):
        return True

    @property
    def is_anonymous(self):
        return False

    def get_full_name(self):
        return '{} {}'.format(self.first_name, self.last_name)

    def __repr__(self):
        return self.get_full_name()

    def __str__(self):
        return self.get_full_name()



from collections import OrderedDict

from norm.executable import NormError
from norm.executable.constant import ListConstant, Constant
from norm.executable.expression import NormExpression
from norm.executable.expression.argument import ArgumentExpr
from norm.executable.expression.condition import ConditionExpr, CombinedConditionExpr
from norm.executable.expression.evaluation import EvaluationExpr, AddDataEvaluationExpr
from norm.grammar.literals import LOP

import logging
logger = logging.getLogger(__name__)


class QueryExpr(NormExpression):

    def __init__(self, op, expr1, expr2):
        """
        Query expression
        :param op: logical operation, e.g., [&, |, ^, =>, <=>]
        :type op: LOP
        :param expr1: left expression
        :type expr1: NormExpression
        :param expr2: right expression
        :type expr2: NormExpression
        """
        super().__init__()
        self.op = op
        self.expr1 = expr1
        self.expr2 = expr2

    def __combine_value(self, value1, value2):
        if value1 is None or value2 is None:
            return None
        elif isinstance(value1, ListConstant) and isinstance(value2, ListConstant):
            value1.value.extend(value2.value)
            return value1
        elif isinstance(value1, ListConstant) and isinstance(value2, Constant):
            value1.value.append(value2.value)
            return value1
        elif isinstance(value1, Constant) and isinstance(value2, ListConstant):
            value2.value.append(value1.value)
            return value2
        elif isinstance(value1, Constant) and isinstance(value2, Constant):
            return ListConstant(value1.type_, [value1.value, value2.value])
        elif isinstance(value1, AddDataEvaluationExpr) and isinstance(value2, AddDataEvaluationExpr):
            # pushdown the combine operation
            if value1.lam is value2.lam:
                combined = self.__combine_data(value1.data, value2.data)
                if combined is not None:
                    value1.data = combined
                    return value1
        return None

    def __combine_data(self, data1, data2):
        cols = list(data1.keys())
        cols.extend(col for col in data2.keys() if col not in data1.keys())
        data = OrderedDict()
        for col in cols:
            combined = self.__combine_value(data1.get(col), data2.get(col))
            if combined is None:
                return None
            else:
                data[col] = combined
        return data

    def compile(self, context):
        if isinstance(self.expr1, AddDataEvaluationExpr) and isinstance(self.expr2, AddDataEvaluationExpr)\
                and self.expr1.lam is self.expr2.lam:
            # Combine the constant arguments for adding data
            # ensure that all arguments are constants
            combined = self.__combine_data(self.expr1.data, self.expr2.data)
            if combined is not None:
                self.expr1.data = combined
                return self.expr1
        if isinstance(self.expr1, ConditionExpr) and isinstance(self.expr2, ConditionExpr):
            return CombinedConditionExpr(self.op, self.expr1, self.expr2).compile(context)
        return self

    def execute(self, context):
        df1 = self.expr1.execute(context)
        df2 = self.expr2.execute(context)
        # TODO: AND to intersect, OR to union
        return df2


class NegatedQueryExpr(NormExpression):

    def __init__(self, expr):
        """
        Negation of the expression
        :param expr: the expression to negate
        :type expr: NormExpression
        """
        super().__init__()
        self.expr = expr

    def compile(self, context):
        if isinstance(self.expr, QueryExpr):
            self.expr.expr1 = NegatedQueryExpr(self.expr.expr1).compile(context)
            self.expr.expr2 = NegatedQueryExpr(self.expr.expr2).compile(context)
            self.expr.op = self.expr.op.negate()
        elif isinstance(self.expr, ConditionExpr):
            self.expr.op = self.expr.op.negate()
        else:
            msg = 'Currently NOT only works on logically combined query or conditional query'
            logger.error(msg)
            raise NotImplementedError(msg)
        return self.expr

    def execute(self, context):
        msg = 'Negated query is not executable, but only compilable'
        logger.error(msg)
        raise NormError(msg)




from norm.executable import NormExecutable
from norm.executable.schema.type import TypeName
from norm.grammar.literals import MOP

import logging
logger = logging.getLogger(__name__)


class Command(NormExecutable):

    def __init__(self, op, type_name):
        """
        The interactive commands.
            * revisions:  show all revisions for the given version or the current version
            * versions: show all historical versions for the given version or all
            * redo: re-apply the next revision
            * undo: roll-back the current revision
        :param op: the operation of the command
        :type op: MOP
        :param type_name: the lambda
        :type type_name: TypeName
        """
        super().__init__()
        self.op: MOP = op
        self.type_name: TypeName = type_name
        from norm.models import Lambda
        self.lam: Lambda = None

    def compile(self, context):
        self.lam = self.type_name.lam
        return self

    def execute(self, context):
        if self.op == MOP.DESCRIBE:
            d = self.lam.data.describe().transpose()
            numerics = set(d.index)
            non_numerics = set(self.lam.data.columns).difference(numerics)
            datetimes = [col for col in non_numerics if self.lam.data[col].dtype.name.find('datetime') >= 0]
            non_numerics = non_numerics.difference(datetimes)
            for col in numerics:
                d.loc[col, 'unique'] = self.lam.data[col].drop_duplicates().count()
            for col in non_numerics:
                d.loc[col] = [self.lam.data[col].count(), None, None, None, None, None, None, None,
                              self.lam.data[col].drop_duplicates().count()]
            for col in datetimes:
                d.loc[col] = [self.lam.data[col].count(), None, None, self.lam.data[col].min(), None, None, None,
                              self.lam.data[col].max(), self.lam.data[col].drop_duplicates().count()]
            for col in self.lam.data.columns:
                d.loc[col, 'type'] = str(self.lam.get_type(col)) or 'Any'
            d['tag'] = ''
            norm_cols = [self.lam.VAR_LABEL, self.lam.VAR_TOMBSTONE, self.lam.VAR_TIMESTAMP, self.lam.VAR_PROB]
            d.loc[norm_cols, 'tag'] = 'reserved'
            for v in self.lam.variables:
                if v.primary:
                    d.loc[v.name, 'tag'] = 'primary'
                else:
                    d.loc[v.name, 'tag'] = 'optional'
            d['count'] = d['count'].astype('int')
            d['unique'] = d['unique'].astype('int')
            return d.loc[self.lam.data.columns, ['tag', 'type', 'count', 'unique', 'min', 'max', 'mean', 'std', '25%',
                                                 '50%', '75%']]
        return None



from norm.executable import NormError, NormExecutable
from norm.models import ListLambda, Lambda, Variable, Status

import logging
logger = logging.getLogger(__name__)


class TypeName(NormExecutable):

    def __init__(self, name, version=None):
        """
        The type qualified name
        :param name: name of the type
        :type name: str
        :param version: version of the type
        :type version: str
        """
        super().__init__()
        self.namespace = None
        self.name = name
        self.version = version
        assert(self.name is not None)
        assert(self.name != '')

    def __str__(self):
        s = self.namespace + '.' if self.namespace else ''
        s += self.name
        s += self.version if self.version is not None else '$latest'
        return s

    def compile(self, context):
        """
        Retrieve the Lambda function by namespace, name, version.
        Note that user is encoded by the version.
        :rtype: Lambda
        """
        if self.namespace is None:
            lam = self.try_retrieve_type(context.session, context.context_namespace, self.name, self.version)
            if lam is None:
                lam = self.try_retrieve_type(context.session, context.search_namespaces, self.name, self.version,
                                             Status.READY)
        else:
            if self.namespace == context.context_namespace:
                lam = self.try_retrieve_type(context.session, self.namespace, self.name, self.version)
            else:
                lam = self.try_retrieve_type(context.session, self.namespace, self.name, self.version, Status.READY)

        self.lam = lam
        return self


class ListType(NormExecutable):

    def __init__(self, intern):
        """
        The type of List with intern type
        :param intern: the type of the intern
        :type intern: TypeName
        """
        super().__init__()
        self.intern = intern

    def compile(self, context):
        """
        Return a list type
        :rtype: ListLambda
        """
        lam = self.intern.lam
        if lam.id is None:
            msg = "{} does not seem to be declared yet".format(self.intern)
            logger.error(msg)
            raise NormError(msg)

        q = context.session.query(ListLambda, Variable).join(ListLambda.variables)\
                           .filter(Variable.type_id == lam.id)
        llam = q.first()
        if llam is None:
            # create a new ListLambda
            llam = ListLambda(lam)
            context.session.add(llam)
        else:
            llam = llam[0]
        self.lam = llam
        return self




"""A collection of ORM sqlalchemy models for managing copyright"""
from norm import config
from norm.models import Model

from sqlalchemy import Column, Integer, String, exists


class License(Model):
    """License for sharing Lambdas"""
    __tablename__ = 'licenses'

    id = Column(Integer, primary_key=True)
    name = Column(String(256), nullable=False)


LICENSES = [
    'MIT',
    'Apache-2.0',
    'CC-0',
    'BSD'
]


def register_licenses():
    for lic in LICENSES:
        in_store = config.session.query(exists().where(License.name == lic)).scalar()
        if not in_store:
            inst = License(name=lic)
            config.session.add(inst)
    config.session.commit()




from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from datetime import datetime
import cachetools
import os
import pandas as pd

# Default namespace stubs
from sqlalchemy.pool import StaticPool

CONTEXT_NAMESPACE_STUB = 'norm.tmp'
USER_NAMESPACE_STUB = 'norm.user'

# Resource control
MAX_LIMIT = 1000000

# Unicode encoding
UNICODE = 'utf-8'

# Version minimal length
VERSION_MIN_LENGTH = 6

# Where the data is stored, e.g., s3://datalake, gs://datalake
NORM_HOME = os.environ.get('NORM_HOME', os.path.expanduser('~/.norm'))
DATA_STORAGE_ROOT = os.environ.get('NORM_DATA_STORAGE_ROOT', os.path.join(NORM_HOME, 'data'))
DB_PATH = os.environ.get('NORM_DB_PATH', os.path.join(NORM_HOME, 'db/norm.db'))

# Cache
cache = cachetools.LRUCache(1024)

# Default user name
PUBLIC_USER = dict(first_name='norm',
                   last_name='ai',
                   username='norm',
                   email='norm@reasoned.ai')

Session = sessionmaker()

# Create database session
try:
    engine = create_engine('sqlite:///{}'.format(DB_PATH), poolclass=StaticPool)
    Session.configure(bind=engine)
    session = Session()
except:
    engine = None
    session = None

# Set the context id
context_id = str(datetime.utcnow().strftime('%m%d%Y.%H%M%S'))

# Pandas display options
pd.options.display.width = 400
pd.options.display.max_columns = 100



from sqlalchemy.ext.declarative import declarative_base

from norm.config import session

import traceback

import logging
logger = logging.getLogger(__name__)

Model = declarative_base()


class Register(object):
    types = []
    types_by_name = {}

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __call__(self, cls):
        self.types.append((cls, self.args, self.kwargs))
        return cls

    @classmethod
    def register(cls):
        for clz, args, kwargs in cls.types:
            if hasattr(clz, 'exists'):
                instance = clz(*args, **kwargs)
                in_store = clz.exists(session, instance)
                if not in_store:
                    logger.info('Registering class {}'.format(instance.name))
                    session.add(instance)
        try:
            session.commit()
        except:
            logger.error('Type registration failed')
            logger.debug(traceback.print_exc())
            session.rollback()

    @classmethod
    def restore_lambda(cls, item):
        if len(cls.types_by_name) == 0:
            for clz, args, kwargs in cls.types:
                instance = clz(*args, **kwargs)
                cls.types_by_name[instance.name] = instance
        instance = cls.types_by_name.get(item)
        if instance is not None and isinstance(instance, Lambda):
            return retrieve_type(instance.namespace, instance.name)
        else:
            return None


class Store(object):

    def __init__(self):
        self._items = dict()

    def __restore(self, item):
        lam = None
        try:
            lam = Register.restore_lambda(item)
            self._items[item] = lam
        except Exception as e:
            msg = 'Norm store is not ready'
            logger.warning(msg)
            logger.warning(e)
        return lam

    def __dir__(self):
        return list(self._items.keys())

    def is_empty(self):
        return len(self._items) == 0

    def __getitem__(self, item):
        lam = self._items.get(item, None)
        if lam is None:
            # reload in case there are more lambdas just got registered
            lam = self.__restore(item)
            if lam is None:
                msg = 'Can not find {}'.format(item)
                logger.warning(msg)

        if lam is not None:
            lam = session.merge(lam)
        return lam

    def __getattr__(self, item):
        return self.__getitem__(item)


from norm.models.norm import (Variable, Lambda, GroupLambda, Status, retrieve_type)
from norm.models.revision import (Revision, revision_variable, SchemaRevision, AddVariableRevision,
                                  DeleteVariableRevision, RenameVariableRevision, RetypeVariableRevision,
                                  DeltaRevision, ConjunctionRevision, DisjunctionRevision, FitRevision)
from norm.models.python import PythonLambda
from norm.models.native import (NativeLambda, TypeLambda, AnyLambda, ListLambda,
                                BooleanLambda, IntegerLambda, StringLambda,
                                PatternLambda, UUIDLambda, FloatLambda,
                                URLLambda, DatetimeLambda)
from norm.models.core import (CoreLambda, StringFormatterLambda, ExtractPatternLambda, ReadFileLambda)
from norm.models.license import License

lambdas = Store()





from norm.executable import NormExecutable
from norm.executable.expression import NormExpression


class SliceExpr(NormExpression):

    def __init__(self, expr, start, end):
        """
        Slice the expression results
        :param expr: the expression to evaluate
        :type expr: NormExpression
        :param start: the start position
        :type start: int
        :param end: the end position
        :type end: int
        """
        super().__init__()
        self.expr: NormExpression = expr
        self.start: int = start
        self.end: int = end

    def compile(self, context):
        self.expr = self.expr.compile(context)
        from norm.executable.schema.variable import VariableName
        from norm.executable.expression.evaluation import EvaluationExpr
        if isinstance(self.expr, VariableName):
            self.expr = EvaluationExpr(args=[], variable=self.expr).compile(context)
        return self

    def execute(self, context):
        df = self.expr.execute(context)
        df = df.iloc[self.start:self.end].reset_index(drop=True)
        # TODO whether reset the index for the projected variable or not?
        return df


class EvaluatedSliceExpr(SliceExpr):

    def __init__(self, expr, expr_range):
        """
        Slice the expression with a range evaluation expression
        :param expr: the expression to slice
        :type expr: NormExpression
        :param expr_range: the expression to evaluate
        :type expr_range: NormExpression
        """
        super().__init__(expr, 0, -1)
        self.expr_range = expr_range



from datetime import datetime
from sqlalchemy import create_engine
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.pool import StaticPool

from norm.engine import NormCompiler, NormError
from norm.config import Session
from norm import config
from norm.security import user

from IPython import get_ipython
from IPython.core.magic import register_line_magic, register_cell_magic, register_line_cell_magic

import os
import logging
logger = logging.getLogger(__name__)

context = NormCompiler(config.context_id, user, config.session)


def configure(home=None, db_path=None, data_path=None, **kwargs):
    """
    Setting the parameters for Norm
    :param home: the path to the db and the data, e.g., /home/<user>/.norm
    :type home: str
    :param db_path: the path to the db, e.g., /home/<user>/.norm/db/norm.db
    :type db_path: str
    :param data_path: the path to the data, e.g., /home/<user>/.norm/data
    :type data_path: str
    :param kwargs: other parameters
    """
    if home is not None:
        config.NORM_HOME = home
        config.DATA_STORAGE_ROOT = os.path.join(home, 'data')
        config.DB_PATH = os.path.join(home, 'db/norm.db')
    if db_path is not None:
        config.DB_PATH = db_path
    if data_path is not None:
        config.DATA_STORAGE_ROOT = data_path
    if home is not None or db_path is not None:
        config.engine = create_engine('sqlite:///{}'.format(config.DB_PATH), poolclass=StaticPool)
        Session.configure(bind=config.engine)
        config.session = Session()
        config.context_id = str(datetime.utcnow().strftime('%m%d%Y.%H%M%S'))
        global context
        context = NormCompiler(config.context_id, user, config.session)


# IPython magics
if get_ipython() is not None:
    @register_line_cell_magic
    def norm(line, cell=None):
        """
        Parsing the norm command and execute it
        :param line: a line of norm command
        :type line: str
        :param cell: a multi-line of norm command
        :type cell: str
        """
        try:
            if cell is None:
                if line.strip('\n\t ')[-1] != ';':
                    line += ';'
                result = context.execute(line)
            else:
                result = context.execute(cell)

            config.session.commit()
            return result
        except SQLAlchemyError as e:
            config.session.rollback()
            msg = 'Session commit failed on {}'.format(config.engine)
            logger.error(msg)
            logger.error(e)
            raise NormError(msg)



import uuid

from norm.executable import NormError
from norm.executable.constant import Constant
from norm.grammar.literals import COP, ConstantType, LOP

from norm.executable.expression import NormExpression
from norm.executable.schema.variable import VariableName

import logging
logger = logging.getLogger(__name__)


class ConditionExpr(NormExpression):

    def __init__(self, op, lexpr, rexpr):
        """
        Condition expression
        :param op: conditional operation, e.g., [<, <=, >, >=, =, !=, in, !in, ~]. ~ means 'like'
        :type op: Union[COP, LOP]
        :param lexpr: arithmetic expression, e.g., a + b - c
        :type lexpr: Union[ArithmeticExpr, ConditionExpr]
        :param rexpr: another arithmetic expression
        :type rexpr: Union[ArithmeticExpr, ConditionExpr]
        """
        super().__init__()
        self.op = op
        self.lexpr = lexpr
        self.rexpr = rexpr
        self._condition = None
        assert(self.lexpr is not None)
        assert(self.rexpr is not None)
        assert(self.op is not None)

    def __str__(self):
        if self._condition is None:
            msg = 'Compile the condition expression first'
            logger.error(msg)
            raise NormError(msg)
        return self._condition

    def compile(self, context):
        if self.op == COP.LK:
            assert(isinstance(self.lexpr, VariableName))
            assert(isinstance(self.rexpr, Constant))
            assert(self.rexpr.type_ is ConstantType.STR or self.rexpr.type_ is ConstantType.PTN)
            self._condition = '{}.str.contains({})'.format(self.lexpr, self.rexpr)
        else:
            self._condition = '({}) {} ({})'.format(self.lexpr, self.op, self.rexpr)
        self.eval_lam = self.lexpr.lam
        from norm.models import Lambda
        self.lam = Lambda(context.context_namespace, context.TMP_VARIABLE_STUB + str(uuid.uuid4()),
                          variables=self.eval_lam.variables)
        self.lam.cloned_from = self.eval_lam
        return self

    def execute(self, context):
        self.lam.data = self.eval_lam.data.query(self._condition, engine='python')
        return self.lam.data


class CombinedConditionExpr(ConditionExpr):

    def __init__(self, op, lexpr, rexpr):
        """
        Combined conditional expression
        :param op: logical operation, e.g., ['and', 'or'] others are not supported yet # TODO support xor, imp, eqv
        :type op: LOP
        :param lexpr: left conditional expression
        :type lexpr: ConditionExpr
        :param rexpr:right conditional expression
        :type rexpr: ConditionExpr
        """
        super().__init__(op, lexpr, rexpr)

    def compile(self, context):
        if self._condition is not None:
            return self

        self.lexpr = self.lexpr.compile(context)
        self.rexpr = self.rexpr.compile(context)
        self._condition = '({}) {} ({})'.format(self.lexpr, self.op, self.rexpr)
        assert(self.lexpr.eval_lam is self.rexpr.eval_lam or self.lexpr.eval_lam is self.rexpr.eval_lam.cloned_from)
        self.eval_lam = self.lexpr.eval_lam
        from norm.models import Lambda
        self.lam = Lambda(context.context_namespace, context.TMP_VARIABLE_STUB + str(uuid.uuid4()),
                          variables=self.eval_lam.variables)
        self.lam.cloned_from = self.eval_lam
        return self

    def execute(self, context):
        self.lam.data = self.eval_lam.data.query(self._condition, engine='python')
        return self.lam.data



from enum import Enum

import logging
logger = logging.getLogger(__name__)

OMMIT = '...'


class CodeMode(Enum):
    QUERY = ''
    PYTHON = '%python'
    KERAS = '%keras'
    SQL = '%sql'


class AOP(Enum):
    ADD = '+'
    SUB = '-'
    MUL = '*'
    DIV = '/'
    MOD = '%'
    EXP = '**'

    def __str__(self):
        return self.value


class COP(Enum):
    GT = '>'
    GE = '>='
    LT = '<'
    LE = '<='
    EQ = '=='
    NE = '!='
    LK = '~'
    IN = 'in'
    NI = 'not in'

    def negate(self):
        if self is COP.GT:
            return COP.LE
        elif self is COP.GE:
            return COP.LT
        elif self is COP.LT:
            return COP.GE
        elif self is COP.LE:
            return COP.GT
        elif self is COP.EQ:
            return COP.NE
        elif self is COP.NE:
            return COP.EQ
        elif self is COP.IN:
            return COP.NI
        elif self is COP.NI:
            return COP.IN
        else:
            msg = 'Negated {} is not supported currently'.format(self)
            logger.error(msg)
            raise NotImplementedError(msg)

    def __str__(self):
        return self.value


class LOP(Enum):
    AND = 'and'
    OR = 'or'
    XOR = 'xor'
    NOT = 'not'
    IMP = 'imp'
    EQV = 'eqv'

    @classmethod
    def parse(cls, token):
        token = token.lower()
        if token == '!' or token == 'not':
            return cls.NOT
        elif token == '&' or token == 'and':
            return cls.AND
        elif token == '|' or token == 'or':
            return cls.OR
        elif token == '^' or token == 'xor':
            return cls.XOR
        elif token == '=>' or token == 'imp':
            return cls.IMP
        elif token == '<=>' or token == '<=>':
            return cls.EQV

    def negate(self):
        if self is LOP.AND:
            return LOP.OR
        elif self is LOP.OR:
            return LOP.AND
        else:
            msg = 'Negation only supports AND/OR for now'
            logger.error(msg)
            raise NotImplementedError(msg)

    def __str__(self):
        if self is LOP.XOR or self is LOP.IMP or self is LOP.EQV:
            msg = 'String representation only supports AND/OR/NOT for now'
            logger.error(msg)
            raise NotImplementedError(msg)
        return self.value


class MOP(Enum):
    HISTORY = 'history'
    DESCRIBE = 'describe'
    REDO = 'redo'
    UNDO = 'undo'
    DEL = 'del'


class ConstantType(Enum):
    NULL = 'none'
    BOOL = 'bool'
    INT = 'integer'
    FLT = 'float'
    STR = 'string'
    PTN = 'pattern'
    UID = 'uuid'
    URL = 'url'
    DTM = 'datetime'
    ANY = 'object'


class ImplType(Enum):
    DEF = ':='
    OR_DEF = '|='
    AND_DEF = '&='



"""A collection of ORM sqlalchemy models for PythonLambda"""
from textwrap import dedent

from pandas import DataFrame, Series
from sqlalchemy import Column, Text, orm

from norm.executable import NormError
from norm.models.norm import Lambda, Status, Variable

import logging
logger = logging.getLogger(__name__)


class PythonLambda(Lambda):

    VAR_INPUTS = 'inputs'

    code = Column(Text, default='')

    __mapper_args__ = {
        'polymorphic_identity': 'lambda_python'
    }

    def __init__(self, namespace, name, description, code):
        """
        Python function, inputs are wrapped into one variable, outputs are one variable too.
        :param namespace: the namespace
        :type namespace: str
        :param name: the name of the function
        :type name: str
        :param description: description
        :type description: str
        :param code: the code of the Python implementation
        :type code: str
        """
        from norm.models import lambdas
        super().__init__(namespace=namespace,
                         name=name,
                         description=description,
                         variables=[Variable(self.VAR_INPUTS, lambdas.Any),
                                    Variable(self.VAR_OUTPUT, lambdas.Any)])
        self.status = Status.READY
        self.code = dedent(code)
        self._load_func()
        self.atomic = True

    def _load_func(self):
        try:
            d = {}
            exec(self.code, d)
            self._func = d.get(self.name)
        except Exception:
            msg = 'Execution errors: \n{}'.format(self.code)
            logger.error(msg)
            raise NormError(msg)
        if self._func is None or not callable(self._func):
            msg = 'Function {} does not exist or failed to parse'.format(self.name)
            logger.error(msg)
            raise NormError(msg)

    @orm.reconstructor
    def init_on_load(self):
        self._load_func()

    def __call__(self, **inputs):
        inp: DataFrame = inputs.get(self.VAR_INPUTS)
        try:
            if inp is not None:
                df = self._func(inp)
                if isinstance(df, Series):
                    df.name = self.VAR_OUTPUT
                if isinstance(inp, DataFrame) and not isinstance(df, DataFrame):
                    raise NormError
            else:
                df = self._func()
        except Exception:
            df = DataFrame(inp.apply(self._func, axis=1), columns=[self.VAR_OUTPUT])
        return df




"""A collection of ORM sqlalchemy models for Revision"""
from sqlalchemy import Column, Integer, String, ForeignKey, Text, orm, JSON
from sqlalchemy import Table
from sqlalchemy.orm import relationship

from norm.models.mixins import ParametrizedMixin
from norm.models.norm import Lambda, Variable
from norm.models import Model

from pandas import DataFrame
import pandas as pd

import logging
logger = logging.getLogger(__name__)

metadata = Model.metadata


class Revision(Model, ParametrizedMixin):
    """Revision of the Lambda. All revisions for the same version are executed in memory."""
    __tablename__ = 'revisions'
    category = Column(String(128))

    PARQUET_EXT = 'parq'

    id = Column(Integer, primary_key=True)
    query = Column(Text, default='')
    position = Column(Integer)
    description = Column(Text, default='')
    lambda_id = Column(Integer, ForeignKey("lambdas.id"))
    lam = relationship(Lambda, back_populates="revisions")

    __mapper_args__ = {
        'polymorphic_identity': 'revision',
        'polymorphic_on': category
    }

    def __init__(self, query, description, lam):
        self.query: str = query
        self.description: str = description
        self.lam: Lambda = lam

    def save(self):
        """
        Save revision information
        """
        raise NotImplementedError

    def apply(self):
        """
        Apply revision on the Lambda
        """
        raise NotImplementedError

    def redo(self):
        """
        Re-apply revision on the Lambda
        """
        raise NotImplementedError

    def undo(self):
        """
        Revert the revision on the Lambda
        """
        raise NotImplementedError


revision_variable = Table(
    'revision_variable', metadata,
    Column('id', Integer, primary_key=True),
    Column('revision_id', Integer, ForeignKey('revisions.id')),
    Column('variable_id', Integer, ForeignKey('variables.id'))
)


class SchemaRevision(Revision):

    __mapper_args__ = {
        'polymorphic_identity': 'revision_schema'
    }

    renames = Column(JSON, default={})
    variables = relationship(Variable, secondary=revision_variable)

    def __init__(self, lam):
        super().__init__('', '', lam)

    def save(self):
        pass

    def redo(self):
        self.apply()


class AddVariableRevision(SchemaRevision):

    __mapper_args__ = {
        'polymorphic_identity': 'revision_schema_add'
    }

    def __init__(self, variables, lam):
        super().__init__(lam)
        current_variable_names = set((v.name for v in self.lam.variables))
        self.variables = [v for v in variables if v.name not in current_variable_names]

    def apply(self):
        self.lam.variables.extend(self.variables)
        for v in self.variables:
            self.lam.data[v.name] = v.type_.default

    def undo(self):
        for v in reversed(self.variables):
            popped_vp = self.lam.variables.pop()
            assert(v.name == popped_vp.name)
        self.lam.data.drop(columns=[v.name for v in self.variables])


class RenameVariableRevision(SchemaRevision):

    __mapper_args__ = {
        'polymorphic_identity': 'revision_schema_rename'
    }

    def __init__(self, renames, lam):
        super().__init__(lam)
        self.renames = renames

    @property
    def renames_r(self):
        return {new_name: old_name for old_name, new_name in self.renames.items()}

    def apply(self):
        if self.lam and self.lam.data is not None:
            self.lam._data = self.lam.data.rename(columns=self.renames)
        for v in self.lam.variables:
            new_name = self.renames.get(v.name)
            if new_name is not None:
                v.name = new_name

    def undo(self):
        renames_r = self.renames_r
        if self.lam and self.lam.data is not None:
            self.lam._data = self.lam.data.rename(columns=renames_r)
        for v in self.lam.variables:
            old_name = renames_r.get(v.name)
            if old_name is not None:
                v.name = old_name


class RetypeVariableRevision(SchemaRevision):

    __mapper_args__ = {
        'polymorphic_identity': 'revision_schema_retype'
    }

    def __init__(self, variables, lam):
        super().__init__(lam)
        current_variable_names = set((v.name for v in self.lam.variables))
        self.variables = [v for v in variables if v.name in current_variable_names]
        variable_names = set((v.name for v in self.variables))
        self.num = len(variable_names)
        assert(len(self.variables) == len(variable_names))
        self.variables.extend([v for v in self.lam.variables if v.name in variable_names])

    @orm.reconstructor
    def init_on_load(self):
        self.num = int(len(self.variables) / 2)

    def apply(self):
        to_variables = self.variables[:self.num]
        retypes = dict((v.name, v.type_) for v in to_variables)
        for v in self.lam.variables:  # type: Variable
            t = retypes.get(v.name)
            if t is not None:
                v.type_ = t

    def undo(self):
        from_variables = self.variables[self.num:]
        retypes = dict((v.name, v.type_) for v in from_variables)
        for v in self.lam.variables:  # type: Variable
            t = retypes.get(v.name)
            if t is not None:
                v.type_ = t


class DeleteVariableRevision(SchemaRevision):

    __mapper_args__ = {
        'polymorphic_identity': 'revision_schema_delete'
    }

    def __init__(self, names, lam):
        super().__init__(lam)
        assert(isinstance(names, list))
        current_variable_names = set((v.name for v in self.lam.variables))
        self.renames = dict((name, '') for name in names if name in current_variable_names)
        self.variables = [v for v in self.lam.variables]

    def apply(self):
        self.lam.variables = [v for v in self.lam.variables if v.name not in self.renames]

    def undo(self):
        self.lam.variables = [v for v in self.lam.variables]


class DeltaRevision(Revision):

    __mapper_args__ = {
        'polymorphic_identity': 'revision_delta'
    }

    def __init__(self, query, description, lam, delta=None):
        super().__init__(query, description, lam)
        self._delta: DataFrame = delta
        self.orig_data: DataFrame = None

    @orm.reconstructor
    def init_on_load(self):
        self._delta = None

    @property
    def path(self):
        return '{}/{}.{}'.format(self.lam.folder, self.id, self.PARQUET_EXT)

    @property
    def delta(self):
        """
        Retrieve the delta. Load from path if not in memory.
        :return: the delta DataFrame
        :rtype: DataFrame
        """
        if self._delta is not None:
            return self._delta

        try:
            self._delta = pd.read_parquet(self.path)
        except FileNotFoundError:
            msg = 'Can not find delta from {}'.format(self.path)
            logger.error(msg)
            raise RuntimeError(msg)
        except:
            self._delta = None
        return self._delta

    @delta.setter
    def delta(self, delta):
        """
        Set the delta if it is not set yet.
        :type delta: DataFrame
        """
        if self._delta is None:
            self._delta = delta

    def save(self):
        if self.delta is None:
            return

        try:
            self._delta.to_parquet(self.path)
        except IOError:
            msg = 'IO problem: can not save delta to {}'.format(self.path)
            logger.error(msg)
            raise
        except:
            msg = 'Other problem: can not save delta to {}'.format(self.path)
            logger.error(msg)
            raise


class ConjunctionRevision(DeltaRevision):
    """
    Revise with conjunction (AND):
        *  If the delta have not been stored (checked by oid), they are appended.
        *  If the delta have been stored as positives (by label), they are overwritten.
        *  If the delta have been stored as negatives, they are ignored.
    """

    __mapper_args__ = {
        'polymorphic_identity': 'revision_delta_conjunction'
    }

    def __init__(self, query, description, lam, delta=None):
        super().__init__(query, description, lam, delta)

    def apply(self):
        self.orig_data = self.lam.data

        oid_col = self.lam.VAR_OID
        label_col = self.lam.VAR_LABEL
        assert(self.lam.data.index.name == oid_col)
        data = self.lam.data
        delta = self.delta
        unstored = delta.index.difference(data.index, sort=False)
        stored = delta.index.intersection(data.index, sort=False)
        if len(unstored) > 0:
            data = data.append(delta.loc[unstored], sort=False)
        if len(stored) > 0:
            overwrites = data.loc[stored][data[label_col] > 0].index
            data.loc[overwrites, delta.columns] = delta.loc[overwrites]
            if label_col not in delta.columns:
                data.loc[overwrites, label_col] = 1.0

        self.lam.data = self.lam.fill_na(data)

    def undo(self):
        self.lam.data = self.orig_data

    def redo(self):
        self.apply()


class DisjunctionRevision(DeltaRevision):
    """
    Revise with disjunction (OR):
        *  If the delta have not been stored (checked by oid), they are appended.
        *  If the delta have been stored as positives (by label), they are ignored.
        *  If the delta have been stored as negatives, they are overwritten.
    """

    __mapper_args__ = {
        'polymorphic_identity': 'revision_delta_disjunction'
    }

    def __init__(self, query, description, lam, delta=None):
        super().__init__(query, description, lam, delta)

    def apply(self):
        self.orig_data = self.lam.data

        oid_col = self.lam.VAR_OID
        label_col = self.lam.VAR_LABEL
        assert(self.lam.data.index.name == oid_col)
        data = self.lam.data
        delta = self.delta
        unstored = delta.index.difference(data.index, sort=False)
        stored = delta.index.intersection(data.index, sort=False)
        if len(unstored) > 0:
            data = data.append(delta.loc[unstored], sort=False)
        if len(stored) > 0:
            overwrites = data.loc[stored][data[label_col] < 1].index
            data.loc[overwrites, delta.columns] = delta.loc[overwrites]
            if label_col not in delta.columns:
                data.loc[overwrites, label_col] = 1.0

        self.lam.data = self.lam.fill_na(data)

    def undo(self):
        self.lam.data = self.orig_data

    def redo(self):
        self.apply()


class FitRevision(DeltaRevision):

    __mapper_args__ = {
        'polymorphic_identity': 'revision_delta_fit'
    }

    def __init__(self, query, description, lam):
        super().__init__(query, description, lam)

    def apply(self):
        pass

    def undo(self):
        pass

    def redo(self):
        pass



from norm.grammar.literals import ConstantType
import datetime
from typing import Union, List


class Constant(object):

    def __init__(self, type_, value=None):
        """
        The constant
        :param type_: the name of the constant type, e.g.,
                      [none, bool, integer, float, string, unicode, pattern, uuid, url, datetime]
        :type type_: ConstantType
        :param value: the value of the constant
        :type value: Union[str, int, float, bool, datetime.datetime, None]
        """
        self.type_: ConstantType = type_
        self.value: Union[str, int, float, bool, datetime.datetime] = value

    def __str__(self):
        if self.type_ in [ConstantType.STR, ConstantType.PTN, ConstantType.UID, ConstantType.URL]:
            return '"{}"'.format(self.value)
        elif self.type_ in [ConstantType.FLT, ConstantType.INT, ConstantType.BOOL]:
            return '{}'.format(self.value)
        elif self.type_ == ConstantType.DTM:
            return self.value.strftime('"%Y-%m-%d %H:%M:%S"')
        else:
            raise NotImplementedError

    def __repr__(self):
        return str(self)

    def execute(self, context):
        return self.value


class ListConstant(Constant):

    def __init__(self, type_, values):
        """
        A list of constant of the same constant type
        :param type_: the name of the constant type
        :type type_: ConstantType
        :param values: the value of the constant
        :type values: List[Union[str, int, float, bool, datetime.datetime]]
        """
        assert(isinstance(values, list))
        super().__init__(type_)
        self.value: List[Union[str, int, float, bool, datetime.datetime]] = values

    def __str__(self):
        return '[' + ','.join(str(v) for v in self.value) + ']'

    def __repr__(self):
        return str(self)



from norm.models.norm import Status, Lambda
from norm.executable import NormExecutable

from typing import Union
import logging
logger = logging.getLogger(__name__)


class VariableName(NormExecutable):

    def __init__(self, scope, name):
        """
        The variable and its scope
        :param scope: the scope of the variable
        :type scope: Union[VariableName, EvaluationExpr]
        :param name: the name of the variable
        :type name: str
        """
        super().__init__()
        from norm.executable.expression.evaluation import EvaluationExpr
        self.scope: Union[VariableName, EvaluationExpr] = scope
        self.name: str = name

    def __str__(self):
        if self.scope is not None:
            return '{}{}{}'.format(self.scope.name, self.VARIABLE_SEPARATOR, self.name)
        else:
            return self.name

    def variable_type(self):
        return self.lam

    def compile(self, context):
        session = context.session
        if self.scope is None:
            name = self.name
            scope = context.get_scope(name)
            if scope is not None:
                return ColumnVariable(scope, name).compile(context)
            else:
                lam = self.try_retrieve_type(session, context.context_namespace, name)
                if lam is None:
                    lam = self.try_retrieve_type(session, context.search_namespaces, name, status=Status.READY)
                self.lam = lam
                return self
        else:
            if isinstance(self.scope, ColumnVariable) and str(self) in self.scope.lam:
                # Already joined
                self.scope.name = str(self)
                return self.scope

            lam = self.scope.variable_type()
            if self.name in lam:
                if isinstance(self.scope, (ColumnVariable, JoinVariable)):
                    # Join lam to the scope for the column
                    return JoinVariable(self.scope, self.name, lam).compile(context)
                else:
                    # A column of the scope variable
                    return ColumnVariable(self.scope, self.name).compile(context)
            else:
                # An evaluation whose first argument is the scope
                lam = self.try_retrieve_type(session, context.context_namespace, self.name)
                if lam is None:
                    lam = self.try_retrieve_type(session, context.search_namespaces, self.name, status=Status.READY)
                assert(lam is not None)
                self.lam = lam
                from norm.executable.expression.argument import ArgumentExpr
                arg = ArgumentExpr(expr=self.scope)
                self.scope = None
                from norm.executable.expression.evaluation import EvaluationExpr
                return EvaluationExpr([arg], self)


class ColumnVariable(VariableName):

    def __init__(self, scope, name):
        super().__init__(scope, name)

    def __str__(self):
        return self.name

    def variable_type(self):
        return self.lam.get_type(self.name)

    def compile(self, context):
        if self.scope is None:
            assert(context.scope is not None)
            self.lam = context.scope
        elif isinstance(self.scope, Lambda):
            self.lam = self.scope
        else:
            self.lam = self.scope.lam
        return self

    def execute(self, context):
        return self.lam.data[self.name]


class JoinVariable(VariableName):

    def __init__(self, scope, name, joiner):
        super().__init__(scope, name)
        self.lam = joiner

    def variable_type(self):
        return self.lam.get_type(self.name)

    def compile(self, context):
        return self

    def execute(self, context):
        lam = self.scope.lam
        joiner = self.lam
        if str(self) not in lam.data.columns:
            to_join = joiner.data[[self.name]].rename(columns={self.name: str(self)})
            lam.data = lam.data.join(to_join, on=str(self.scope))
        return lam.data

